@@ -1,3 +1,31 @@
+"""
+The top-level script you write using pydatatask should call `pydatatask.main.main` in its ``if __name__ == '__main__'``
+block. This will parse ``sys.argv`` and display the administration interface for the pipeline.
+
+The help screen should look something like this:
+
+.. code::
+
+    $ python3 main.py --help
+      usage: main.py [-h] {update,run,status,trace,rm,ls,cat,inject,launch,shell} ...
+
+    positional arguments:
+      {update,run,status,trace,rm,ls,cat,inject,launch,shell}
+        update              Keep the pipeline in motion
+        run                 Run update in a loop until everything is quiet
+        status              View the pipeline status
+        trace               Track a job's progress through the pipeline
+        rm                  Delete data from the pipeline
+        ls                  List jobs in a repository
+        cat                 Print data from a repository
+        inject              Dump data into a repository
+        launch              Manually start a task
+        shell               Launch an interactive shell to interrogate the pipeline
+
+    options:
+      -h, --help            show this help message and exit
+"""
+
 from typing import Callable, Dict, List, Optional, Set, Union
 import argparse
 import asyncio
@@ -8,11 +36,9 @@ import sys
 import IPython
 import yaml
 
-import pydatatask
-
 from .pipeline import Pipeline
 from .repository import BlobRepository, MetadataRepository, Repository
-from .task import Task, settings
+from .task import Task
 
 l = logging.getLogger(__name__)
 token_re = re.compile(r"\w+\.\w+")
@@ -25,16 +51,25 @@ __all__ = (
     "delete_data",
     "inject_data",
     "print_status",
+    "print_trace",
     "launch",
     "shell",
     "run",
 )
 
+# pylint: disable=missing-function-docstring,missing-class-docstring
+
 
 def main(
     pipeline: Pipeline,
     instrument: Optional[Callable[[argparse._SubParsersAction], None]] = None,
 ):
+    """
+    The pydatatask main function! Call this with the pipeline you've constructed to parse ``sys.argv`` and display the
+    pipeline administration interface.
+
+    If you like, you can pass as the ``instrument`` argument a function which will add additional commands to the menu.
+    """
     parser = argparse.ArgumentParser()
     subparsers = parser.add_subparsers(dest=argparse.SUPPRESS, required=True)
 
@@ -137,6 +172,7 @@ async def main_inner(pipeline, coro):
 
 
 def shell(pipeline: Pipeline):
+    pydatatask = __import__("pydatatask")
     assert pipeline
     assert pydatatask
     IPython.embed(using="asyncio")
@@ -157,33 +193,54 @@ async def run(pipeline: Pipeline, forever: bool, launch_once: bool, timeout: Opt
             raise TimeoutError("Pipeline run timeout")
 
 
+def get_repos(pipeline: Pipeline, all_repos: bool):
+    seen = set()
+    for task in pipeline.tasks.values():
+        for link in task.links.values():
+            if not all_repos and not link.is_status and not link.is_input and not link.is_output:
+                continue
+            if link.repo in seen:
+                continue
+            seen.add(link.repo)
+            yield link.repo
+
+
 async def print_status(pipeline: Pipeline, all_repos: bool):
-    for task in pipeline.tasks.values():  # TODO parallelize - waiting on repos to be a top-level abstraction
+    async def inner(repo: Repository):
+        the_sum = 0
+        async for _ in repo:
+            the_sum += 1
+        return the_sum
+
+    repo_list = list(get_repos(pipeline, all_repos))
+    repo_sizes = dict(zip(repo_list, await asyncio.gather(*(inner(repo) for repo in repo_list))))
+
+    for task in pipeline.tasks.values():
         print(task.name)
         for link_name, link in sorted(
             task.links.items(),
             key=lambda x: 0 if x[1].is_input else 1 if x[1].is_status else 2,
         ):
-            if not all_repos and not link.is_status and not link.is_input and not link.is_output:
-                continue
-            the_sum = 0
-            async for _ in link.repo:
-                the_sum += 1
-            print(f"{task.name}.{link_name} {the_sum}")
+            if link.repo in repo_sizes:
+                print(f"{task.name}.{link_name} {repo_sizes[link.repo]}")
         print()
 
 
 async def print_trace(pipeline: Pipeline, all_repos: bool, job: List[str]):
-    for task in pipeline.tasks.values():  # TODO parallelize - waiting on repos to be a top-level abstraction
+    async def inner(repo: Repository):
+        return [j for j in job if await repo.contains(j)]
+
+    repo_list = list(get_repos(pipeline, all_repos))
+    repo_members = dict(zip(repo_list, await asyncio.gather(*(inner(repo) for repo in repo_list))))
+
+    for task in pipeline.tasks.values():
         print(task.name)
         for link_name, link in sorted(
             task.links.items(),
             key=lambda x: 0 if x[1].is_input else 1 if x[1].is_status else 2,
         ):
-            if not all_repos and not link.is_status and not link.is_input and not link.is_output:
-                continue
-            jobs = [j for j in job if await link.repo.contains(j)]
-            print(f'{task.name}.{link_name} {" ".join(jobs)}')
+            if link.repo in repo_members:
+                print(f"{task.name}.{link_name} {' '.join(repo_members[link.repo])}")
         print()
 
 
@@ -204,11 +261,11 @@ async def delete_data(pipeline: Pipeline, data: str, recursive: bool, job: List[
                 jobs = job
                 check = True
 
-            async def del_job(j):
-                await dependant.delete(j)
-                print(j, dependant)
+            async def del_job(j, dep):
+                await dep.delete(j)
+                print(j, dep)
 
-            await asyncio.gather(*[del_job(j) for j in jobs if not check or await dependant.contains(j)])
+            await asyncio.gather(*[del_job(j, dependant) for j in jobs if not check or await dependant.contains(j)])
 
 
 async def list_data(pipeline: Pipeline, data: List[str]):
@@ -218,7 +275,7 @@ async def list_data(pipeline: Pipeline, data: List[str]):
     for token in tokens:
         task, repo = token.split(".")
         await namespace[task].consume(repo)
-    result = eval(input_text, {}, namespace)
+    result = eval(input_text, {}, namespace)  # pylint: disable=eval-used
 
     for job in sorted(result):
         print(job)
@@ -264,7 +321,7 @@ async def inject_data(pipeline: Pipeline, data: str, job: str):
     item = pipeline.tasks[taskname].links[reponame].repo
 
     if isinstance(item, BlobRepository):
-        with item.open(job, "wb") as fp:
+        async with await item.open(job, "wb") as fp:
             while True:
                 data_bytes = await asyncio.get_running_loop().run_in_executor(None, sys.stdin.buffer.read, 1024 * 1024)
                 if not data:
@@ -284,8 +341,7 @@ async def inject_data(pipeline: Pipeline, data: str, job: str):
 
 async def launch(pipeline: Pipeline, task_name: str, job: str, sync: bool, meta: bool, force: bool):
     task = pipeline.tasks[task_name]
-    job = job
-    settings(sync, meta)
+    pipeline.settings(sync, meta)
 
     if force or await task.ready.contains(job):
         await task.launch(job)
